package logger

import (
	"context"
	"encoding/json"
	"fmt"
	"net/http"
	"sync"

	"github.com/replicate/cog/pkg/model"
	"github.com/replicate/cog/pkg/util/console"
)

type MessageType string

const (
	MessageTypeLogLine   MessageType = "log_line"
	MessageTypeDebugLine MessageType = "debug_line"
	MessageTypeError     MessageType = "error"
	MessageTypeStatus    MessageType = "status"
	MessageTypeVersion   MessageType = "version"
)

type Message struct {
	Type    MessageType    `json:"type"`
	Text    string         `json:"data"`
	Version *model.Version `json:"version"`
}

type StreamLogger struct {
	ctx    context.Context
	writer http.ResponseWriter
	mu     sync.Mutex
}

func NewStreamLogger(ctx context.Context, w http.ResponseWriter) *StreamLogger {
	return &StreamLogger{ctx: ctx, writer: w}
}

func (logger *StreamLogger) logText(messageType MessageType, text string) {
	msg := Message{Type: messageType, Text: text}
	data, err := json.Marshal(msg)
	if err != nil {
		console.Warnf("Failed to marshal console text: %s", text)
	}
	logger.write(data)
}

func (logger *StreamLogger) write(data []byte) {
	// don't write if the request has been cancelled
	select {
	case <-logger.ctx.Done():
		return
	default:
	}

	data = append(data, '\n')
	logger.mu.Lock()
	defer logger.mu.Unlock()
	if _, err := logger.writer.Write(data); err != nil {
		console.Warnf("HTTP response writer failed to write: %s", data)
		return
	}
	if f, ok := logger.writer.(http.Flusher); ok {
		f.Flush()
	} else {
		console.Warn("HTTP response writer can not be flushed")
	}
}

func (logger *StreamLogger) Info(line string) {
	logger.logText(MessageTypeLogLine, line)
}

func (logger *StreamLogger) Debug(line string) {
	logger.logText(MessageTypeDebugLine, line)
}

func (logger *StreamLogger) Infof(line string, args ...interface{}) {
	logger.logText(MessageTypeLogLine, fmt.Sprintf(line, args...))
}

func (logger *StreamLogger) Debugf(line string, args ...interface{}) {
	logger.logText(MessageTypeDebugLine, fmt.Sprintf(line, args...))
}

func (logger *StreamLogger) WriteStatus(status string, args ...interface{}) {
	logger.logText(MessageTypeStatus, fmt.Sprintf(status, args...))
}

func (logger *StreamLogger) WriteError(err error) {
	logger.logText(MessageTypeError, err.Error())
}

func (logger *StreamLogger) WriteVersion(version *model.Version) {
	msg := Message{Type: MessageTypeVersion, Version: version}
	data, err := json.Marshal(msg)
	if err != nil {
		console.Warn("Failed to marshal version")
	}
	logger.write(data)
}
